from typing import Type, Union

import gym
import numpy as np
import torch as th

from ddpg import DDPG
from env_wrapper import EnvWrapper
from networks import DDPGPolicy


class HER(object):
    """
    Hindsight Experience Replay (HER)

    :param policy: (DDPGPolicy) The policy model to use. (MlpPolicy)
    :param env: (gym.GoalEnv) The environment to learn from.
    :param model: (DDPG) Off policy model which will be used with hindsight experience replay. (DDPG)
    :param n_goals: (int) Number of sampled goals for replay.
    :param goal_strategy: (str) Strategy for sampling goals for replay. One of ['episode', 'final', 'future']
    """

    def __init__(self, policy: Type[DDPGPolicy], env: gym.GoalEnv, model: Type[DDPG], n_goals: int = 5,
                 goal_strategy: str = 'final', *args,
                 **kwargs):
        super(HER, self).__init__()

        self.policy_class = policy
        self.env = EnvWrapper(env)
        self.model = model(self.policy_class, self.env, *args, **kwargs)

        # storage for transitions of current episode
        self.episode_storage = []
        self.n_goals = n_goals
        self.goal_strategy = goal_strategy.lower()

    def sample_goals(self, sample_idx: int) -> Union[np.ndarray, None]:
        """
        Sample a goal based on goal_strategy.

        :param sample_idx: (int) Index of current transition.
        :return: (np.ndarray or None) Return sampled goal.
        """
        if self.goal_strategy == 'final':
            # replay with final state of current episode
            return self.episode_storage[-1][0]['achieved_goal']
        elif self.goal_strategy == 'future':
            # replay with random state which comes from the same episode and was observed after current transition
            # we have no transition after last transition of episode
            if (sample_idx + 1) < len(self.episode_storage):
                index = np.random.choice(np.arange(sample_idx + 1, len(self.episode_storage)))
                return self.episode_storage[index][0]['achieved_goal']
        elif self.goal_strategy == 'episode':
            # replay with random state which comes from the same episode as current transition
            index = np.random.choice(np.arange(len(self.episode_storage)))
            return self.episode_storage[index][0]['achieved_goal']
        else:
            raise ValueError('Strategy for sampling goals not supported!')

    def store_transitions(self) -> None:
        """
        Store current episode in replay buffer. Sample additional goals and store new transitions in replay buffer.
        """

        # iterate over current episodes transitions
        for idx, trans in enumerate(self.episode_storage):

            observation, action, reward, new_observation, done = trans

            # concatenate observation with (desired) goal
            obs = np.concatenate([observation['observation'], observation['desired_goal']])
            new_obs = np.concatenate([new_observation['observation'], new_observation['desired_goal']])

            # store data in replay buffer
            self.model.replay_buffer.add(obs, new_obs, action, reward, done)

            # sample set of additional goals
            sampled_goals = [sample for sample in (self.sample_goals(idx) for i in range(self.n_goals)) if
                             sample is not None]

            # iterate over sampled goals and store new transitions in replay buffer
            for goal in sampled_goals:
                # compute new reward with new goal
                new_reward = self.env.compute_reward(achieved_goal=new_observation['achieved_goal'], desired_goal=goal,
                                                     info=None)

                # concatenate observation with (desired) goal
                obs = np.concatenate([observation['observation'], goal])
                new_obs = np.concatenate([new_observation['observation'], goal])

                # store data in replay buffer
                self.model.replay_buffer.add(obs, new_obs, action, new_reward, done)

    def learn(self, num_epochs: int, num_episodes: int = 20, gradient_steps: int = 40) -> 'HER':
        """
        Perform training loop.

        :param num_epochs: (int) Number of epochs.
        :param num_episodes: (int) Number of total episodes per epoch.
        :param gradient_steps: (int) Number of gradient steps after num_episodes.
        """
        num_timesteps = 0
        observation = self.env.reset()

        for epoch in range(num_epochs):
            for episode in range(num_episodes):

                done = False
                episode_reward, episode_timesteps = 0.0, 0

                while not done:

                    # concatenate observation and (desired) goal
                    obs = np.concatenate([observation['observation'], observation['desired_goal']])

                    # Select action randomly or according to policy
                    if num_timesteps < self.model.learning_starts:
                        # Warmup phase
                        unscaled_action = np.array([self.model.action_space.sample()])
                        # Rescale the action from [low, high] to [-1, 1]
                        scaled_action = self.model.scale_action(unscaled_action)
                    else:
                        with th.no_grad():
                            # Convert to pytorch
                            obs_ = th.FloatTensor(np.array(obs).reshape(1, -1)).to(self.model.device)
                            scaled_action = self.model.policy.predict(obs_).numpy()

                    # Add noise to the action
                    clipped_action = scaled_action
                    if self.model.action_noise is not None:
                        clipped_action = np.clip(clipped_action + self.model.action_noise(), -1, 1)

                    # Rescale and perform action
                    new_observation, reward, done, info = self.env.step(self.model.unscale_action(clipped_action)[0])

                    episode_reward += reward

                    # add current transition to episode storage
                    self.episode_storage.append((observation, clipped_action, reward, new_observation, done))

                    observation = new_observation

                    num_timesteps += 1
                    episode_timesteps += 1

                if done:
                    # store episode in replay buffer
                    self.store_transitions()
                    # clear storage for current episode
                    self.episode_storage = []

                    if self.model.action_noise is not None:
                        self.model.action_noise.reset()

                    observation = self.env.reset()

                    # Display training infos
                    if self.model.verbose >= 1:
                        print(
                            f"Epoch: {epoch + 1} - episode: {episode + 1} "
                            f"- timestep: {num_timesteps} - Reward = {episode_reward:.2f}")
            # Train
            if num_timesteps > 0 and num_timesteps > self.model.learning_starts:
                self.model.train(gradient_steps, batch_size=self.model.batch_size)

        return self
